#!/usr/bin/env python3


"""
┌───────┐ ┌───────┐ ┌───────┐ ┌───────┐
│ A     │ │ A     │ │ A     │ │ A     │
│       │ │       │ │       │ │       │
│   ♧   │ │   ♤  │ │   ♥   │ │   ♢   │
│       │ │       │ │       │ │       │
│     A │ │     A │ │     A │ │     A │
└───────┘ └───────┘ └───────┘ └───────┘

Welkom!
Dit is een introductie voor sorteer-algoritmes.
Bovendien kunt u iets leren over objectgeörienteerd programeren.
Helemaal beneden staat een gebruiksvoorbeeld.


                               --- OPDRACHT: ---
Schrijf er een functie om de lijst van kaarten te sorteren, maar
    ZONDER ingebouwde functies te gebruiken (bv. "sorted()" of "<list>.sort()")!
                               -----------------

Tips:
- Kaarten hebben een "beeld" en een "nummer"
    (zie variables "beelden" en "nummers" hieronder).
- Denk er eerst over na hoe je echte kaarten zou sorteren.
- Een kaartenspel/kaartenlijst is zoals een gewone lijst.
- Misschien gaan dus de volgende functies helpen:
    <list>.append(<Kaart>)
    <list>.index()
    <list>.reverse()
    <list>.insert(<positie>)
    <list>.pop(<positie>)
    (https://www.w3schools.com/python/python_ref_list.asp)
- U mag hier in het bestand op het einde toevoegen,
    of dit als bibliotheek importeren in een andere bestand met
    import Kaarten
    speelkaartenlijst = Kaarten.Kaartspel()

Succes!


Bug-meldingen graag aan <mielke-ws@mailbox.org>. Dank u!
"""

#________________________________________________________________________________
# bibliotheken importeren
import numpy.random as RND # numpy "random" voor toevalsprocessen (schudden)


#________________________________________________________________________________
# hier gaan we alle beeldjes en nummers versamelen
beelden = ["klaver", "schop", "hart", "ruit"]

nummers = [f"{nr}" for nr in range(2,11)] \
        + [ "boer", "vrouw", "heer", "ace" \
        # , "joker" \
        ]

# We bewaren nog symbolletjes om de kaarten overzichtelijk te kunnen afprinten.
symbolen = { \
              "klaver": "♧" \
            , "schop": "♤" \
            , "hart": "♥" \
            , "ruit": "♢" \
            , "boer": "B" \
            , "vrouw": "V" \
            , "heer": "R" \
            , "ace": "A" \
            , "joker": "J" \
            }

#________________________________________________________________________________
### Objekt: Kaart
# we defineren hier een "class"!
# Dit hoort bij "objectgeoriënteerd programmeren" (OOP),
#     zie https://nl.wikipedia.org/wiki/Objectgeori%C3%ABnteerd
# In de taalrijke commentaren staat er meer over uitgelegd.
class Kaart(object):
    # Hier volgt de bouwplan voor een kaart, het houdt alles in wat een kaart
    #     moet hebben.
    # Dit kunnen dus variables zijn, of functies.
    # Zoals met de definitie van functies gaat de definitie van een kaart door
    #     indentatie bepaald zijn.
    # Zoals bij functies wordt eigenlijk nog niet ineens een "kaart" aangemakt.
    # Pas als we het object oproepen (bv. "Kaart('hart', 'boer')"), dan maken
    #     we een kaart met specifieke eigenschappen aan.

    def __init__(self, beeld, nummer):
        # Deze functie dient voor de initialisatie;
        # ze wordt telkens opgeroepen als we een concrete kaart aanmaken.
        # We noemen het een "CONSTRUCTOR"!
        #
        # De variable "self" houdt het object zelf bij, will zeggen dat we met
        # "self" op de eigenschappen van een specifieke kaart kunnen toegrijpen.
        #    "self" moet er bij elke functie binnen een object het eerste
        #    argument zijn.


        # input verifiëren
        if beeld not in beelden:
            raise IOError("Kies één van de beelden, aub: " + ", ".join(beelden) )

        if nummer not in nummers:
            raise IOError("Kies één van de nummers, aub: " + ", ".join(nummers) )

        # zet de eigenschappen
        self.beeld = beeld
        self.nummer = nummer

        # kleur (is hier niet belangrijk)
        self.kleur = "rood" if beeld in ['hart', 'ruit'] else "zwart"


    def Afbeelding(self):
        # hiermee kunnen we een kaart mooi afdrukken.
        Extend = lambda symb: " "+symb if len(symb) == 1 else symb
        nr = Extend(symbolen.get(self.nummer, self.nummer))
        bld = Extend(symbolen.get(self.beeld, self.beeld))

        # terug geven we een ASCII-afbeelding van de kaart.
        return f"""
 ┌───────┐
 │{nr}     │
 │       │
 │  {bld}   │
 │       │
 │    {nr} │
 └───────┘"""


    def Symbool(self):
        # ... of we kunnen het symbol teruggeven, als afkorting.
        nr = symbolen.get(self.nummer, self.nummer)
        bld = symbolen.get(self.beeld, self.beeld)

        return bld+nr


    def __str__(self):
        # deze "magische functie" bepaald welke tekst eruit komt als we die
        # kaart in een string omvormen
        return f"{self.beeld} {self.nummer}"

#________________________________________________________________________________
# en nog een object: het KAARTSPEL
# ... een versameling van kaarten.
class Kaartspel(list):
    # Zoals boven maken we hier een object aan.
    # We gebruiken hier "overerving":
    #    Een Kaartspel is in feite een lijst.
    #    We kunnen er alles mee doen wat we met lijsten doen.
    #    Maar er komen nog functies bij.

    def __init__(self, schudden = True):
        # De "CONSTRUCTOR" van een kaartspel.
        # We maken hier dus een nieuw spel aan,
        # met alle combinaties van beeld/nummer die we hebben.
        # Normaal wordt het bij de initialisatie geschud.

        # eerst gaan we alle kaartjes aanmaken
        kaartjes = [Kaart(bld, nr) \
                 for bld in beelden \
                 for nr in nummers \
                 if bld not in ["joker"] \
                 ]

        # We gaan nu een lijst aanmaken.
        # "Kaartspel" is in feite een lijst (overerving).
        # we willen dus de constructor van "list" oproepen met alle kaartjes als
        #     argumenten.
        # "list" is de SUPERCLASS van "Kaartspel", darom gebruiken we de
        #    functie "super()"!
        super(Kaartspel, self).__init__(kaartjes)
        # ^- dit roept dus de SUPERCONSTRUCTOR van het Kaartspel op.

        # print("sorted 1: ", IsSorted(self))

        # Het kaartspel is nu aangemaakt.
        # We gaan het nog schudden!
        self.Schudden()


    def Schudden(self, verbose = False):
        # Hier gaan we de kaartjes schudden.

        # Eerst moeten we de kaartjes in een nieuwe lijst opslaan.
        kaartjes = [kaart for kaart in self]

        # We gaan een geschudde lijst met kaartnummers aanmaken.
        geschudde_indices = RND.choice(  range(len(self)), size = len(self) \
                                       , replace = False)

        # we gaan dit vermelden indien de gebruiker het wenst.
        if verbose:
            print ("we schudden...")

        # en nu gaan we de kaart één voor één uitwisselen.
        for nieuw_idx, oude_idx in enumerate(geschudde_indices):
            self[nieuw_idx] = kaartjes[oude_idx]


    def TrekToevallig(self):
        # hier gaan we één kaart trekken/opnemen vanuit het set.
        return self.pop(RND.choice(range(len(self))))


    def Afdrukken(self):
        # met die functie kunnen we het deck afdrukken
        print ([kaart.Symbool() for kaart in self])

    def __str__(self):
        return f"Een kaartspel van {len(self)} kaarten."

#________________________________________________________________________________
# is het kaartspel gesorteerd?
def IsSorted(kaartenlijst):
    # Deze functie geeft "True" terug als de kaartenlijst gesorteerd is.
    # Dit betekent, minimaal, ofwel:
    #  - de beelden zijn in groepjes
    #  - en de nummers overal in dezelfde volgorde
    # of:
    #  - de nummers zijn in groepjes
    #  - en de beelden overal in dezelfde volgorde
    # Beelden en nummers noem ik "eigenschappen" van een kaart.

    # we moeten twee eigenschappen checken
    GetBeeld = lambda kaart: kaart.beeld
    GetNummer = lambda kaart: kaart.nummer
    getters = [GetBeeld, GetNummer]

    # (I) kijken of de kaarten naar beeld gesorteerd zijn
    # (II) kijken of nummers veranderd zijn
    eigenschappen = {}
    wanordes = {}
    for iteratie, Eigenschap in enumerate(getters):
        veranderingen = sum([ \
                        Eigenschap(kaartenlijst[idx-1]) \
                              != Eigenschap(kaartenlijst[idx]) \
                        for idx in range(1, len(kaartenlijst)) \
                        ])

        # eigenschappen opslaan
        eigenschappen[iteratie] = set([Eigenschap(krt) for krt in kaartenlijst])

        # in een beeld-gesorteerde lijst moet het aantal veranderingen één kleiner
        # zijn dan het totaal aantal beelden
        wanordes[iteratie] = veranderingen >= len(eigenschappen[iteratie])

    if all(wanordes.values()):
        # geen van de eigenschappen is in de juiste volgorde
        return False

    # één van de eigenschappen (beeld of nummer) is oké, we moeten de andere bekijken.
    for iteratie, wanorde in wanordes.items():
        if wanorde:
            continue

        # spelen met indices :)
        andere_iteratie = not iteratie
        Eigenschap = getters[iteratie]
        AndereEigenschap = getters[andere_iteratie]

        # we maken er een dict van met lijsten van de andere eigenschap
        eigenschap_lijsten = {eig: [ AndereEigenschap(kaart) \
                                     for kaart in kaartenlijst \
                                     if Eigenschap(kaart) == eig \
                                    ] \
                              for eig in eigenschappen[iteratie]}

        # We moeten eerst kijken of alle elementen in de lijsten
        #     parallel hetzelfde zijn.
        # We "zippen" de lijsten,
        #     dan kijken we dat er maar één en hetzelfde andere eigenschap is.
        for checks in zip(*eigenschap_lijsten.values()):
            if not len(set(checks)) == 1:
                return False

    # als we tot nu toe nog alles in orde hebben, is de lijst gesorteerd! :)
    return True


#________________________________________________________________________________
if __name__ == "__main__":
    # deze code gaat het programma laten draaien als het _niet_ als bibliotheek
    # gebruikt wordt.

    # als voorbeeld maken we een hele kaartspel aan.
    speelkaartenlijst = Kaartspel()
    print (speelkaartenlijst)

    print ("Is de kaartlijst gesorteerd? " \
           , "Ja!" if IsSorted(speelkaartenlijst) else "Nee!" \
           )

    # om alle kaarten te zien hebben we boven een functie aangemaakt
    speelkaartenlijst.Afdrukken()

    # we trekken er één kaart van.
    kaart = speelkaartenlijst.TrekToevallig()
    print("we hebben de volgende kaart getrokken: \n"+kaart.Afbeelding())
    print (f"nu zijn er nog {len(speelkaartenlijst)} kaartjes over.")


#________________________________________________________________________________
# end of file.
